% $Id: DesugarTest.oz,v 1.10 2011/04/15 15:53:36 leavens Exp leavens $
% AUTHOR: Gary T. Leavens

\insert 'Desugar.oz'
\insert 'Unparse.oz'
\insert 'TestingNoStop.oz' %% from the course library
declare

% A helping procedure to eliminate repitition in the testing below
proc {DSTest Expected AST}
   %% ENSURES: Desugaring and the expected output are shown, and compared
   UPAST = {VirtualString.toString {Unparse AST}}
in
   {System.showInfo ""}
   {System.showInfo "Desugar(" # UPAST # ")"}
   local
      StrExpected = {VirtualString.toString Expected}
      Ans = {Desugar AST}
      UPAns = {Unparse Ans}
      Same = {VirtualString.toString UPAns} == StrExpected
   in
      {System.showInfo "       == " # UPAns}
      if {Not Same}
      then
         {System.showInfo "Expected: " # StrExpected}
      end
      {Assert Same}
   end
end

{StartTesting 'Desugar'}

{DSTest "skip"
 skipStmt
}

{DSTest "local A in skip end"
 localStmt("A" skipStmt)
}

{DSTest "A = B"
 assignStmt("A" varIdExp("B"))
}

{DSTest "A = 3"
 assignStmt("A" numExp(3))
}

{DSTest "A = 3 B = C"
 seqStmt([assignStmt("A" numExp(3)) assignStmt("B" varIdExp("C"))])
}

{DSTest "A = 3 B = C E = A"
 seqStmt([assignStmt("A" numExp(3)) assignStmt("B" varIdExp("C"))
          assignStmt("E" varIdExp("A"))])
}

{DSTest "if B then skip else skip end"
 ifStmt(varIdExp("B") skipStmt skipStmt)
}

{DSTest "local Unnest in Unnest = true if Unnest then skip else skip end end"
 ifStmt(boolExp(true) skipStmt skipStmt)
}

{DSTest "local Unnest in Unnest = true if Unnest then R = true else S = false end end"
 ifStmt(boolExp(true)
        assignStmt("R" boolExp(true))
        assignStmt("S" boolExp(false)))
}

{DSTest "if X then A = B else skip end"
 ifStmt(varIdExp("X") assignStmt("A" varIdExp("B")) skipStmt)
}

{DSTest "if X then A = B else D = Z end"
 ifStmt(varIdExp("X")
        assignStmt("A" varIdExp("B"))
        assignStmt("D" varIdExp("Z")))
}

{DSTest "local A in A = B end"
 localStmt("A" assignStmt("A" varIdExp("B")))
}

{DSTest "local X in X = C skip end"
 caseStmt(varIdExp("C") varIdPat("X") skipStmt skipStmt)
}

{DSTest "case C of ok then skip else skip end"
 caseStmt(varIdExp("C") atomPat(ok) skipStmt skipStmt)
}

{DSTest "local X in X = C B = X end"
 caseStmt(varIdExp("C") varIdPat("X")
          assignStmt("B" varIdExp("X"))
          skipStmt)
}

{DSTest "local X in X = C B = X end"
 caseStmt(varIdExp("C") varIdPat("X")
          assignStmt("B" varIdExp("X"))
          assignStmt("D" numExp(7)))
}

{DSTest "case C of nowXisFree then B = X else D = 7 end"
 caseStmt(varIdExp("C") atomPat(nowXisFree)
          assignStmt("B" varIdExp("X"))
          assignStmt("D" numExp(7)))
}

{DSTest "local Q in Q = C B = X end"
 caseStmt(varIdExp("C")
          varIdPat("Q")
          assignStmt("B" varIdExp("X"))
          assignStmt("Z" varIdExp("Q")))
}

{DSTest "case C of foo() then B = X else D = 7 end"
 caseStmt(varIdExp("C") recordPat(foo nil)
          assignStmt("B" varIdExp("X"))
          assignStmt("D" numExp(7)))
}

{DSTest "case C of foo(X Y) then B = X else D = Y end"
 caseStmt(varIdExp("C") recordPat(foo [posFld(varIdExp("X"))
                                       posFld(varIdExp("Y"))])
          assignStmt("B" varIdExp("X"))
          assignStmt("D" varIdExp("Y")))
}

{DSTest "case C of foo(f1:X f2:Y) then B = X else D = Y end"
 caseStmt(varIdExp("C") recordPat(foo [colonFld(f1 varIdExp("X"))
                                       colonFld(f2 varIdExp("Y"))])
          assignStmt("B" varIdExp("X"))
          assignStmt("D" varIdExp("Y")))
}

{DSTest "case C of foo(f1:CasePat)"
     # " then case CasePat of bar(f2:X f3:Y f4:Z)"
     # " then B = X if Z then D = Y else skip end else skip end"
     # " else skip end"
 caseStmt(varIdExp("C")
          recordPat(foo
                    [colonFld(f1
                              recordExp(atomExp(bar)
                                        [colonFld(f2 varIdExp("X"))
                                         colonFld(f3 varIdExp("Y"))
                                         colonFld(f4 varIdExp("Z"))]))])
          seqStmt([assignStmt("B" varIdExp("X"))
                  ifStmt(varIdExp("Z") assignStmt("D" varIdExp("Y")) skipStmt)])
          skipStmt)
}

{DSTest "case C of foo(f1:CasePat)"
     # " then case CasePat of bar(f2:X f3:Y f4:Q)"
     # " then B = X if Z then D = Y else skip end else skip end"
     # " else skip end"
 caseStmt(varIdExp("C")
          recordPat(foo
                    [colonFld(f1
                              recordExp(atomExp(bar)
                                        [colonFld(f2 varIdExp("X"))
                                         colonFld(f3 varIdExp("Y"))
                                         colonFld(f4 varIdExp("Q"))]))])
          seqStmt([assignStmt("B" varIdExp("X"))
                  ifStmt(varIdExp("Z") assignStmt("D" varIdExp("Y")) skipStmt)])
          skipStmt)
}

{DSTest "{P}"
 applyStmt(varIdExp("P") nil)
}

{DSTest "{Map Ls F}"
 applyStmt(varIdExp("Map") [varIdExp("Ls") varIdExp("F")])
}

{DSTest "local Unnest in Unnest = 3 {Browse Unnest} end"
 applyStmt(varIdExp("Browse") [numExp(3)])
}


{DSTest "local Unnest in"
 # " local Unnest1 in"
 # " Unnest1 = 3"
 # " local Unnest2 in"
 # " Unnest2 = nil"
 # " Unnest = Unnest1|Unnest2"
 # " end"
 # " end"
 # " {Map Unnest F}"
 # " end"
 applyStmt(varIdExp("Map")
           [recordExp(atomExp('|') [posFld(numExp(3)) posFld(atomExp(nil))])
            varIdExp("F")])
}

{DSTest
 "local Unnest in"
 # " local Unnest1 in"
 # " Unnest1 = 3"
 # " local Unnest2 in"
 # " Unnest2 = nil"
 # " Unnest = Unnest1|Unnest2"
 # " end"
 # " end"
 # " local Unnest1 in"
 # " Unnest1 = proc {$ X R} {Browse X} R = X end"
 # " {Map Unnest Unnest1}"
 # " end"
 # " end"
 applyStmt(varIdExp("Map")
           [recordExp(atomExp('|') [posFld(numExp(3)) posFld(atomExp(nil))])
            procExp([varIdPat("X") varIdPat("R")]
                    seqStmt([applyStmt(varIdExp("Browse") [varIdExp("X")])
                             assignStmt("R" varIdExp("X"))]))])
}

{DSTest "Three = proc {$ R} R = 3 end"
 assignStmt("Three" procExp([varIdPat("R")] assignStmt("R" numExp(3))))
}

{DSTest "Three = proc {$ Result} Result = 3 end"
 namedFunStmt("Three" nil numExp(3))
}

{DSTest "Add = proc {$ X Y Result} {Plus X Y Result} end"
 namedFunStmt("Add" [varIdPat("X") varIdPat("Y")]
              applyExp(varIdExp("Plus") [varIdExp("X") varIdExp("Y")]))
}

{DSTest "Add = proc {$ X Y Result} {Plus X X Z Result} end"
 namedFunStmt("Add" [varIdPat("X") varIdPat("Y")]
              applyExp(varIdExp("Plus") [varIdExp("X") varIdExp("X")
                                         varIdExp("Z")]))
}

{DSTest "F = proc {$ X Result} Result = proc {$ F R} {F A R} end end"
 namedFunStmt("F" [varIdPat("X")]
              procExp([varIdPat("F") varIdPat("R")]
                      assignStmt("R" applyExp(varIdExp("F") [varIdExp("A")]))))
}

{DSTest
 "InTest = proc {$ Tree R}"
 # " local X in local Y in"
 # " Tree = node(X Y) {Plus X Y R}"
 # " end end end"
 assignStmt("InTest"
            procExp([varIdPat("Tree") varIdPat("R")]
                    inStmt(recordPat(node
                                     [posFld(varIdExp("X"))
                                      posFld(varIdExp("Y"))])
                           varIdExp("Tree")
                           assignStmt("R"
                                      applyExp(varIdExp("Plus")
                                               [varIdExp("X")
                                                varIdExp("Y")])))))
}

{DSTest
 "local Unnest in"
 # " Unnest = noElse"
 # " {Exception.raise Unnest}"
 # " end"
 applyStmt(varIdExp("Exception.raise") [atomExp('noElse')])
 }

{DSTest
 "local Unnest in local Unnest1 in"
 # " Unnest1 = noElse()"
 # " Unnest = kernel(Unnest1)"
 # " end"
 # " {Exception.raise Unnest}"
 # " end"
 applyStmt(varIdExp("Exception.raise")
                   [recordExp(atomExp('kernel')
                              [posFld(recordExp(atomExp('noElse')
                                                nil))])])
 }

{DSTest "First = proc {$ ArgPat Result} case ArgPat of X#Y then Result = X else local Unnest in Unnest = noElse {Exception.raise Unnest} end end end"
 namedFunStmt("First" [recordPat('#' [posFld(varIdExp("X"))
                                      posFld(varIdExp("Y"))])]
              varIdExp("X"))
}

{DSTest "RevPair = proc {$ ArgPat Result} case ArgPat of X#Y then Result = Y#X else local Unnest in Unnest = noElse {Exception.raise Unnest} end end end"
 namedFunStmt("RevPair" [recordPat('#' [posFld(varIdExp("X"))
                                      posFld(varIdExp("Y"))])]
              recordExp(atomExp('#')
                        [posFld(varIdExp("Y")) posFld(varIdExp("X"))]))
}

{DSTest "RevPair = proc {$ P Result} case P of X#Y then Result = Y#X else Result = error end end"
 namedFunStmt("RevPair" [varIdPat("P")]
              caseExp(varIdExp("P")
                      recordPat('#' [posFld(varIdExp("X"))
                                     posFld(varIdExp("Y"))])
                      recordExp(atomExp('#')
                                [posFld(varIdExp("Y")) posFld(varIdExp("X"))])
                      atomExp(error)))
}

{DSTest
 "T = proc {$ N Result}"
 # " local Unnest in {Odd N Unnest}"
 # " if Unnest"
 # " then local Unnest1 in"
 # " local Unnest2 in"
 # " local Unnest3 in"
 # " Unnest3 = 3"
 # " {Mult Unnest3 N Unnest2}"
 # " end"
 # " local Unnest3 in"
 # " Unnest3 = 1"
 # " {Plus Unnest2 Unnest3 Unnest1}"
 # " end"
 # " end"
 # " local Unnest2 in"
 # " Unnest2 = 2"
 # " {Div Unnest1 Unnest2 Result}"
 # " end"
 # " end"
 # " else"
 # " local Unnest1 in"
 # " Unnest1 = 2"
 # " {Div N Unnest1 Result}"
 # " end"
 # " end"
 # " end"
 # " end"
  namedFunStmt("T" [varIdPat("N")]
              ifExp(applyExp(varIdExp("Odd") [varIdExp("N")])
                    applyExp(varIdExp("Div")
                             [applyExp(varIdExp("Plus")
                                       [applyExp(varIdExp("Mult")
                                                 [numExp(3) varIdExp("N")])
                                        numExp(1)])
                              numExp(2)])
                    applyExp(varIdExp("Div") [varIdExp("N") numExp(2)])))
}

local
   ParsedFigExample = % (Figure 2 of homework 3a)
   namedFunStmt("AddToEach"
                [recordPat('#'
                           [posFld(varIdExp("A")) posFld(varIdExp("B"))])
                 varIdPat("Ls")]
      caseExp(varIdExp("Ls")
              recordPat('|'
                        [posFld(recordExp(atomExp('#')
                                         [posFld(varIdExp("X"))
                                          posFld(varIdExp("Y"))]))
                         posFld(varIdExp("T"))])
              recordExp(atomExp('|')
                        [posFld(recordExp(atomExp('#')
                                          [posFld(applyExp(varIdExp("Plus") 
                                                           [varIdExp("A")
                                                            varIdExp("X")]))
                                           posFld(applyExp(varIdExp("Plus")
                                                           [varIdExp("B")
                                                            varIdExp("Y")]))]))
                         posFld(applyExp(varIdExp("AddToEach")
                                         [recordExp(atomExp('#')
                                                    [posFld(varIdExp("A"))
                                                     posFld(varIdExp("B"))])
                                          varIdExp("T")]))])
              atomExp(nil)))
in
   {DSTest
    "AddToEach = proc {$ ArgPat Ls Result}"
    # " case ArgPat of A#B then"
    # " case Ls of CasePat|T then" 
    # " case CasePat of X#Y then"
    # " local Unnest in"
    # " local Unnest1 in"
    # " {Plus A X Unnest1}"
    # " local Unnest2 in"
    # " {Plus B Y Unnest2}"
    # " Unnest = Unnest1#Unnest2"
    # " end"
    # " end"
    # " local Unnest1 in"
    # " local Unnest2 in"
    # " Unnest2 = A#B"
    # " {AddToEach Unnest2 T Unnest1}"
    # " end"
    # " Result = Unnest|Unnest1"
    # " end"
    # " end"
    # " else"
    # " Result = nil"
    # " end"
    # " else Result = nil"
    # " end"
    # " else local Unnest in Unnest = noElse {Exception.raise Unnest} end"
    # " end"
    # " end"
    ParsedFigExample
   }
end

{DSTest "thread skip end"
 threadStmt(skipStmt)}

{DSTest "thread local Unnest in Unnest = 3 local X in Unnest = X {P X} end end end"
 threadStmt(inStmt(varIdPat("X") numExp(3) applyStmt(varIdExp("P") [varIdExp("X")])))}

{DSTest "thread X = 7 end"
 assignStmt("X" threadExp(numExp(7)))}

{DSTest "local Unnest in thread Unnest = 7 end {P Unnest} end"
 applyStmt(varIdExp("P") [threadExp(numExp(7))])}

{StartTesting done}
